# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)
from os import unlink

import MDAnalysis as mda

from aquaduct.utils import clui
from aquaduct.utils.helpers import create_tmpfile


class TmpDumpWriterOfMDA(object):
    def __init__(self):

        # creates tmp file
        self.pdbfile = create_tmpfile(ext='pdb')

        # create mda writer
        self.mdawriter = mda.Writer(self.pdbfile, multiframe=True)

    def dump_frames(self, reader, frames, selection='protein'):

        to_dump = reader.parse_selection(selection)

        for frame in frames:
            if frame < reader.number_of_frames:
                reader.set_real_frame(frame)
                self.mdawriter.write(to_dump)
            else:
                logger.error(
                    'Requested frame %d exceeded available number of frames %d.' % (frame, reader.number_of_frames - 1))

    def close(self):
        self.mdawriter.close()
        return self.pdbfile

    def __del__(self):
        unlink(self.pdbfile)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides functions for clusterization.
Clusterization is done by :mod:`scikit-learn` module.
"""

import numpy as np
from sklearn.cluster import Birch, DBSCAN, AffinityPropagation, KMeans, MeanShift, estimate_bandwidth
from scipy.spatial.distance import cdist

# problems with clustering methods and size of set
# DBSCAN:              n > 0
# AffinityPropagation: n > 0
# KMeans:              n > clusters
# MeanShift:           n > 6


AVAILABLE_METHODS = ['dbscan','kmeans', 'affprop', 'meanshift', 'birch','barber']

def get_required_params(method):
    if method == 'kmeans':
        return ['n_clusters']


from aquaduct.utils.helpers import Auto
from aquaduct.utils import clui

class BarberClusterResult(object):

    def __init__(self,labels_):
        self.labels_ = np.array(labels_)

class BarberCluster(object):

    def fit(self,coords,radii=None):
        friends = {}
        for nr,(coord,radius) in enumerate(zip(coords,radii)):
            distances = cdist([coord],coords,metric='euclidean').flatten() - np.array(radii) - radius
            # less then zero are intersecting
            if (distances<=0).any():
                friends.update({nr:np.argwhere(distances<=0).flatten().tolist()})
        # loop over friends' groups
        clustered = []
        clusters = []
        for leader,group in friends.iteritems():
            if leader in clustered: continue
            cluster = [] + group
            last_size = -1
            while len(cluster) != last_size:
                last_size = len(cluster)
                for leader2,group2 in friends.iteritems():
                    if leader2 in clustered: continue
                    if set(cluster).intersection(set(group2)):
                        cluster += group2
                        cluster = list(set(cluster))
            clustered += cluster
            clusters.append(cluster)
        clusters.sort(key=lambda x: len(x),reverse=True)
        # make labels
        labels = [None]*len(coords)
        for nr,cluster in enumerate(clusters):
            for c in cluster:
                labels[c] = nr
        return BarberClusterResult(labels)


def MeanShiftBandwidth(X, **kwargs):
    if 'bandwidth' in kwargs:
        if kwargs['bandwidth'] is Auto:
            bandwidth = estimate_bandwidth(np.array(X), quantile=0.5)  # TODO: change it to the default value of 0.3 or use it as option?
            kwargs.update({'bandwidth': bandwidth})
            clui.message("Meanshift automatic bandwidth calculation: bandwidth = %f" % float(
                bandwidth))  # TODO: make it properly
    return kwargs


class PerformClustering(object):
    # aqeuduct clustering helper class

    def __init__(self, method, **kwargs):

        self.method = method
        self.method_kwargs = kwargs
        self.method_results = None
        self.clusters = None

    def __str__(self):
        out = str(self.method.__name__)
        return out

    def __call__(self, coords, radii=None):
        # compatibility
        return self.fit(coords, radii=radii)

    def _get_noclusters(self, n):
        return [0] * n

    def fit(self, coords, radii=None):
        # radii are used for Barber only
        if len(coords) < 2:
            self.clusters = self._get_noclusters(len(coords))
            return self.clusters
        # special cases
        if self.method is BarberCluster:
            method = self.method()
            self.method_results = method.fit(coords,radii)
            self.clusters = map(int, self.method_results.labels_ + 1)
        else:
            if self.method is MeanShift:
                if len(coords) < 6:
                    self.clusters = self._get_noclusters(len(coords))
                    clui.message("Number of objects %d < 6 is too small for MeanShift method. Skipping." % (len(coords)))
                    return self.clusters
                method = self.method(**MeanShiftBandwidth(coords, **self.method_kwargs))
            else:
                if self.method is KMeans:
                    if 'n_clusters' in self.method_kwargs:
                        if len(coords) < self.method_kwargs['n_clusters']:
                            self.clusters = self._get_noclusters(len(coords))
                            clui.message(
                                "Number of objects %d < %d is too small for KMeans method. Skipping." % (
                                    len(coords), self.method_kwargs['n_clusters']))
                            return self.clusters
                method = self.method(**self.method_kwargs)
            self.method_results = method.fit(coords)
            self.clusters = map(int, self.method_results.labels_ + 1)
        return self.clusters

    def centers(self):
        assert self.clusters is not None, "Perform clusterization first."
        assert self.method_results is not None, "Perform clusterization first."

        if hasattr(self.method_results, 'cluster_centers_'):
            return self.method_results.cluster_centers_
        raise NotImplementedError('Cluster centers is not implemented for %r method yet.' % self.method)




# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from os.path import splitext

import MDAnalysis as mda
from MDAnalysis.topology.core import guess_atom_element

from aquaduct.traj.selections import SelectionMDA

mda_available_formats = {re.compile('(nc|NC)'): 'nc',
                         re.compile('(parmtop|top|PARMTOP|TOP)'): 'parmtop',
                         re.compile('(dcd|DCD)'): 'LAMMPS',
                         re.compile('(psf|PSF)'): 'psf',
                         re.compile('(pdb|PDB)'): 'pdb',
                         re.compile('(xtc|XTC)'): 'XTC'}


class Reader(object):
    def __init__(self, topology, trajectory, window=None):
        assert isinstance(topology, str)
        if not isinstance(trajectory, str):
            for trj in trajectory:
                assert isinstance(trj, str)

        self.topology_file_name = topology
        self.trajectory_file_name = trajectory

        self.trajectory_object = self.open_trajectory()

        # if window is set then behaviour of iteration over frames related methods is different
        self.frames_window = window
        self.frames_window_list = None

        # check window
        lo = min(self.get_start_frame(),self.get_stop_frame())
        hi = max(self.get_start_frame(), self.get_stop_frame())
        assert lo >=0 , 'Wrong frames window definition, wrong low bound %d; should be >= 0' % lo
        assert hi <= self.real_number_of_frames, 'Wrong frames window definition, wrong high bound %d; should be <= %d' % (hi,self.real_number_of_frames-1)

        self.set_real_frame(0)  # ensure that by default it starts from frame 0


    def open_trajectory(self):
        # should return any object that can be further used to parse trajectory
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    @property
    def real_number_of_frames(self):
        # should return real number of frames
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def next_frame(self):
        # should set next frame or raise StopIteration
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def set_real_frame(self, frame):
        # should set current frame according to window
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    ###########################################################################
    # window params

    def get_start_frame(self):
        if self.frames_window is None:
            return 0
        if self.frames_window.start is None:
            return 0
        return self.frames_window.start

    def get_step_frame(self):
        if self.frames_window is None:
            return 1
        if self.frames_window.step is None:
            return 1
        return self.frames_window.step

    def get_stop_frame(self):
        if self.frames_window is None:
            return self.real_number_of_frames - 1
        if self.frames_window.stop is None:
            return self.real_number_of_frames - 1
        return self.frames_window.stop

    def get_window_frame_range(self):
        return xrange(self.get_start_frame(),self.get_stop_frame()+1,self.get_step_frame())
        #return xrange(self.get_start_frame(),self.get_stop_frame(),self.get_step_frame())

    ###########################################################################
    # window dependent

    @property
    def number_of_frames(self):
        if self.frames_window is None:
            return self.real_number_of_frames
        start = self.get_start_frame()
        stop = self.get_stop_frame()
        step = self.get_step_frame()
        return (abs(stop-start)+1)/step

    def iterate_over_frames(self):
        # should return list of frames ids or generator returning such a list, and should set appropriate frame
        # appropriate frame means current frame that can be recalculated to real frame (window)
        for current_frame,real_frame in enumerate(self.get_window_frame_range()):
            self.set_real_frame(real_frame)
            yield current_frame

    def cf2rf(self,cf):
        # current frame to real frame
        if self.frames_window_list is None:
            self.frames_window_list = list(self.get_window_frame_range())
        return self.frames_window_list[cf]

    def set_current_frame(self, frame):
        # frame is current frame, derrive real frame
        return self.set_real_frame(self.cf2rf(frame))

    ###########################################################################

    def parse_selection(self, selection):
        # should parse and return selection object
        # is this object updated automatically depends on the particular implementation of Reader class
        # in particular MDA updates postions of atoms accoriding to current frame, to renew selection one has to parse it again
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def select_resnum(self, resnum):
        # should return selection object
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")

    def select_multiple_resnum(self, resnum_list):
        # should return selection object
        raise NotImplementedError("This is abstract class. Missing implementation in a child class.")


class ReadViaMDA(Reader):

    ############################################################################
    # window dependent


    ############################################################################

    def set_real_frame(self, frame):
        self.trajectory_object.trajectory[frame]

    @property
    def real_number_of_frames(self):
        # should return number of frames
        return len(self.trajectory_object.trajectory)

    def next_frame(self):
        try:
            self.trajectory_object.trajectory.next()
        except:
            raise StopIteration

    def parse_selection(self, selection):
        return SelectionMDA(self.trajectory_object.select_atoms(selection))

    def select_resnum(self, resnum):
        assert isinstance(resnum, (int, long))
        return self.parse_selection("resnum %d" % resnum)

    def select_multiple_resnum(self, resnum_list):
        assert isinstance(resnum_list, list)
        selection = None
        for resnum in resnum_list:
            if selection is None:
                selection = self.select_resnum(resnum)
            else:
                selection += self.select_resnum(resnum)
        return SelectionMDA(selection)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        self.trajectory_object.trajectory.close()

    def open_trajectory(self):
        topology = splitext(self.topology_file_name)[1][1:]
        for afk in mda_available_formats.keys():
            if afk.match(topology):
                topology = mda_available_formats[afk]
                break
        trajectory = splitext(self.trajectory_file_name[0])[1][1:]
        for afk in mda_available_formats.keys():
            if afk.match(trajectory):
                trajectory = mda_available_formats[afk]
                break
        return mda.Universe(self.topology_file_name,
                            self.trajectory_file_name,
                            topology_format=topology,
                            format=trajectory)


class ReadAmberNetCDFviaMDA(ReadViaMDA):
    def open_trajectory(self):
        return mda.Universe(self.topology_file_name,
                            self.trajectory_file_name,
                            topology_format="prmtop", format="nc")


class ReadDCDviaMDA(ReadViaMDA):
    def open_trajectory(self):
        return mda.Universe(self.topology_file_name,
                            self.trajectory_file_name,
                            topology_format="psf", format="dcd")



VdW_radii = {'ac': 2.47,
             'ag': 2.11,
             'al': 1.84,
             'am': 2.44,
             'ar': 1.88,
             'as': 1.85,
             'at': 2.02,
             'au': 2.14,
             'b': 1.92,
             'ba': 2.68,
             'be': 1.53,
             'bi': 2.07,
             'bk': 2.44,
             'br': 1.85,
             'c': 1.7,
             'ca': 2.31,
             'cd': 2.18,
             'ce': 2.42,
             'cf': 2.45,
             'cl': 1.75,
             'cm': 2.45,
             'co': 2.0,
             'cr': 2.06,
             'cs': 3.43,
             'cu': 1.96,
             'dy': 2.31,
             'er': 2.29,
             'es': 2.45,
             'eu': 2.35,
             'f': 1.47,
             'fe': 2.04,
             'fm': 2.45,
             'fr': 3.48,
             'ga': 1.87,
             'gd': 2.34,
             'ge': 2.11,
             'h': 1.1,
             'he': 1.4,
             'hf': 2.23,
             'hg': 2.23,
             'ho': 2.3,
             'i': 1.98,
             'in': 1.93,
             'ir': 2.13,
             'k': 2.75,
             'kr': 2.02,
             'la': 2.43,
             'li': 1.82,
             'lr': 2.46,
             'lu': 2.24,
             'md': 2.46,
             'mg': 1.73,
             'mn': 2.05,
             'mo': 2.17,
             'n': 1.55,
             'na': 2.27,
             'nb': 2.18,
             'nd': 2.39,
             'ne': 1.54,
             'ni': 1.97,
             'no': 2.46,
             'np': 2.39,
             'o': 1.52,
             'os': 2.16,
             'p': 1.8,
             'pa': 2.43,
             'pb': 2.02,
             'pd': 2.1,
             'pm': 2.38,
             'po': 1.97,
             'pr': 2.4,
             'pt': 2.13,
             'pu': 2.43,
             'ra': 2.83,
             'rb': 3.03,
             're': 2.16,
             'rh': 2.1,
             'rn': 2.2,
             'ru': 2.13,
             's': 1.8,
             'sb': 2.06,
             'sc': 2.15,
             'se': 1.9,
             'si': 2.1,
             'sm': 2.36,
             'sn': 2.17,
             'sr': 2.49,
             'ta': 2.22,
             'tb': 2.33,
             'tc': 2.16,
             'te': 2.06,
             'th': 2.45,
             'ti': 2.11,
             'tl': 1.96,
             'tm': 2.27,
             'u': 2.41,
             'v': 2.07,
             'w': 2.18,
             'xe': 2.16,
             'y': 2.32,
             'yb': 2.26,
             'zn': 2.01,
             'zr': 2.23}
'''
Dictionary of VdW radii.

Data taken from L. M. Mentel, mendeleev, 2014. Available at: https://bitbucket.org/lukaszmentel/mendeleev.
Package **mendeleev** is not used because it depends on too many other libraries.
'''

def atom2vdw_radius(atom):
    '''
    Function tries to guess atom element and checks if it is in :attr:`VdW_radii` dictionary. If it fails 1.4 is returned.
    Guessing is done twice:

    #. Function :func:`MDAnalysis.topology.core.guess_atom_element` is used.
    #. :attr:`MDAnalysis.core.AtomGroup.Atom.element` is used.

    :param MDAnalysis.core.AtomGroup.Atom atom: Atom of interest.
    :rtype: float
    :return: VdW radius.
    '''
    element = str(guess_atom_element(atom.name)).lower()
    if element in VdW_radii:
        return VdW_radii[element]
    element = str(atom.element).lower()
    if element in VdW_radii:
        return VdW_radii[element]
    return 1.4



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# this modlue is a prototype and have to be rewritten

import logging
logger = logging.getLogger(__name__)

import multiprocessing
from multiprocessing import Queue, Manager, Lock, Value, Process
from itertools import izip_longest
from functools import partial

import numpy as np
from scipy.spatial.distance import cdist, pdist

from aquaduct.traj.paths import GenericPathTypeCodes, GenericPaths, yield_single_paths, MasterPath
from aquaduct.utils.helpers import list_blocks_to_slices, strech_zip, zip_zip, xzip_xzip, concatenate
from aquaduct.utils import clui
from aquaduct.utils.maths import make_default_array, defaults
from aquaduct.traj.inlets import InletClusterGenericType, InletClusterExtendedType

################################################################################
part2type_dict = {0: GenericPathTypeCodes.scope_name,
                  1: GenericPathTypeCodes.object_name,
                  2: GenericPathTypeCodes.scope_name}
'''
Part number to :class:`~aquaduct.traj.paths.GenericPathTypeCodes` dictionary.
'''

parts = (0, 1, 2)
'''
Parts enumerate.
'''

################################################################################


class CTypeSpathsCollectionWorker(object):
    '''
    Worker class for averaging spaths in points of master path.
    '''
    def __init__(self, spaths=None, ctype=None, bias_long=5, smooth=None):
        '''
        Core method for averaging spaths in to master path.

        Averaging is done in chunks.

        :param list spaths: List of separate paths to average.
        :param InletClusterGenericType ctype: CType of spaths.
        :param int bias_long: Bias towards long paths used in :meth:`lens_norm`.
        :param Smooth smooth: Smoothing method.
        '''

        self.spaths = spaths
        assert isinstance(ctype, InletClusterGenericType) or isinstance(ctype, InletClusterExtendedType)
        self.ctype = ctype
        self.bias_long = bias_long # TODO: check if it is required here
        self.smooth = smooth

        self.lens_cache = None
        self.lens_real_cache = None
        self.lens_norm_cache = None
        self.full_size_cache = None

    def coords_types_prob_widths(self, sp_slices_):
        '''
        Calculates average coordinates, type and width in given chunk.

        Parameter :attr:`sp_slices_` is tuple of length equal to number of spaths. It contains slices for all spaths respectively. With these slices spaths are cut and **only** resulting chunks are used for calculations.

        Therefore, this method average spaths in one point of master math. This point is defined by slices submitted as :attr:`sp_lices_` parameter.

        Algorithm of averaging (within current chunks of spaths):

        #. Coordinates for all spaths are collected.
        #. Lengths of all spaths are collected (from cached variables) and kept as lists of lengths equal to chunks' sizes.

            .. note::

                Lengths of collected lengths of spaths are of the same size as coordinates

        #. New coordinates are calculated as weighted average of collected coordintates with :func:`numpy.average`. As weights collected lengths are used.

            .. note::

                Function :func:`numpy.average` is called with flatten coordinates and lengths.

        #. Width of average path is calculated as mean value of flatten coordinates mutual distances.
        #. Type of average paths is calculated as probability (frequency) of :attr:`~aquaduct.traj.paths.GenericPathTypeCodes.scope_name`.

        :param tuple sp_slices_: Slices that cut chunks from all paths.
        :rtype: 3 element tuple
        :return: coordinates, type (frequency), and width of averaged spaths in current point
        '''

        # get zz coords, zz means zip_zip - for all spaths
        coords_zz = [sp.get_coords_cont(smooth=self.smooth)[sl] for sp, sl in zip(self.spaths, sp_slices_)]

        # make lens_zz which are lens corrected to the lenghts of coords_zz and normalized to zip_zip number of obejcts
        lens_zz = []
        for l, coord_z in zip(self.lens_cache, coords_zz):
            # l is lenght for one spath
            # coord_z are coordinates of this path (sliced to current chunk)
            if len(coord_z) > 0:
                lens_zz.append([float(l) / len(coord_z)] * len(coord_z))  # normalize and correct lengths
            else:
                # lens_zz.append([float(l)] * len(coord_z))
                lens_zz.append([])

        # here we have coords_zz, lens_zz
        # and we can calculate coords, types_prob, widths

        # concatenate zip_zip coords and lens
        coords_zz_cat = list(concatenate(*coords_zz))
        del coords_zz

        lens_zz_cat = list(concatenate(*lens_zz))
        del lens_zz
        # average coords_zz_cat using weights of lens_zz_cat

        coords_to_append = make_default_array(np.average(coords_zz_cat, axis=0, weights=lens_zz_cat))
        del lens_zz_cat

        # calculate widths
        if len(self.spaths) > 1:  # is the len of coords_zz the same as sp_slices_ and self.spaths?
            widths_to_append = make_default_array(np.mean(pdist(coords_zz_cat, 'euclidean')))
        else:
            widths_to_append = 0.
        del coords_zz_cat

        # concatenate zip_zip gtypes
        types_zz_cat = list(concatenate(*[sp.gtypes_cont[sl] for sp, sl in zip(self.spaths, sp_slices_)]))
        del sp_slices_
        # append type porbability to types

        types_to_append = float(types_zz_cat.count(GenericPathTypeCodes.scope_name)) / len(types_zz_cat)

        return coords_to_append, types_to_append, widths_to_append

    def __call__(self, nr_sp_slices_):
        '''
        Callable interface.

        :param tuple nr_sp_slices_: Two element tuple: nr and sp_slice
        '''
        return nr_sp_slices_[0], self.coords_types_prob_widths(nr_sp_slices_[-1])


class CTypeSpathsCollection(object):
    '''
    Object for grouping separate paths that belong to the same CType.
    Method :meth:`get_master_path` allows for calculation of average path.
    '''
    parts = (0, 1, 2)  # spath parts
    '''
    Enumeration of spath parts.
    '''

    # takes group of paths belonging to one ctype and allows to get MasterPath
    def __init__(self, spaths=None, ctype=None, bias_long=5, pbar=None, threads=1):
        '''
        :param list spaths: List of separate paths.
        :param InletClusterGenericType ctype: CType of spaths.
        :param int bias_long: Bias towards long paths used in :meth:`lens_norm`.
        :param pbar: Progress bar object.
        :param int threads: Number of available threads.
        '''
        self.pbar = pbar
        self.threads = threads
        logger.debug("Threads passed %d", threads)

        self.spaths = spaths
        assert isinstance(ctype, InletClusterGenericType) or isinstance(ctype, InletClusterExtendedType)
        self.ctype = ctype
        self.bias_long = bias_long

        # precompute some values
        self.beat()

        with clui.tictoc('spaths props cache in %s' % str(self.ctype)):
            self.lens_cache = self.lens()
            self.lens_real_cache = self.lens_real()
            self.lens_norm_cache = self.lens_norm()
            self.full_size_cache = self.full_size()
        self.beat()

        # self.lock = Lock()

    def beat(self):
        '''
        Touch progress bar, if any.
        '''
        if self.pbar is not None:
            self.pbar.heartbeat()

    def update(self):
        '''
        Update progres bar by one, if any.
        '''
        if self.pbar is not None:
            self.pbar.next()

    def lens(self):
        '''
        Returns total lengths of all paths.

        If ctype in #:# and not 0 and not None then take length of `object` part only.

        :return: Total (or `object` part) lengths of all paths.
        :rtype: numpy.ndarray
        '''
        if self.ctype.input is not None:
            if self.ctype.input > 0:
                if self.ctype.input == self.ctype.output:
                    return make_default_array([float(len(sp.types_object)) for sp in self.spaths])
        return make_default_array([float(sp.size) for sp in self.spaths])

    def lens_norm(self):
        '''
        Returns normalized lengths calculated by :meth:`lens`.

        Applied normalization is twofold:

        #. All lengths are divided by maximal length, and
        #. All lengths are subjected to :func:`pow` function with p = :attr:`bias_long`.

        :return: Normalized total (or `object` part) lengths of all paths.
        :rtype: numpy.ndarray
        '''
        lens = self.lens()
        if np.max(lens) > 0:
            lens /= np.max(lens)  # normalize
            return lens ** self.bias_long  # bias to long paths

    def lens_real(self):
        '''
        Returns real lengths of all paths.

        :return: Sizes of all paths.
        :rtype: list
        '''
        return [sp.size for sp in self.spaths]

    def full_size(self):
        '''
        Returns desired size of master path.

        :return: Size of master path.
        :rtype: int
        '''
        # first check what is the size of paths in all parts and normalize and then scale them
        sizes = []
        for part in self.parts:
            # lengths of all paths of part part
            lens = make_default_array([float(len(sp.types[part])) for sp in self.spaths])
            if np.max(lens) > 0:
                lens /= np.max(lens)  # normalization
                lens = lens ** self.bias_long  # scale them by increasing weights of long paths
            if sum(lens) == 0:
                sizes.append(0)
            else:
                # weighted average by paths lengths
                sizes.append(int(np.average([len(sp.types[part]) for sp in self.spaths], 0, lens)))
        return sum(sizes)  # total size (desired)

    @staticmethod
    def simple_types_distribution(types):
        '''
        Calculates normalized sizes of incoming, object, and outgoing parts of spath using generic types.

        It is assumed that spath has object part.

        :param list types: List of generic types.
        :rtype: 3 element list
        :return: Normalized sizes of incomin, object, and outgoing parts.
        '''
        # possible types are:
        # GenericPathTypeCodes.object_name
        # GenericPathTypeCodes.scope_name
        td_in, td_obj, td_out = 0, 0, 0
        sls = list(list_blocks_to_slices(types))
        if GenericPathTypeCodes.scope_name in types[sls[0]]:
            # this is input part
            td_in = len(types[sls[0]])
        if GenericPathTypeCodes.scope_name in types[sls[-1]]:
            # this is output part
            td_out = len(types[sls[-1]])
        # the rest is object
        td_obj = len(types) - td_in - td_out
        return map(lambda x: float(x) / len(types), (td_in, td_obj, td_out))

    def types_distribution(self):
        '''
        :rtype: numpy.matrix
        :return: median values of :meth:`simple_types_distribution` for all spaths.
        '''
        # make median distribuitions
        return np.matrix(make_default_array(np.median([self.simple_types_distribution(sp.gtypes_cont) for sp in self.spaths], axis=0)))

    def types_prob_to_types(self, types_prob):
        '''
        Changes types probabilities as returned by :meth:`CTypeSpathsCollectionWorker.coords_types_prob_widths` to types.

        :param list types_prob: List of types probabilities.
        :rtype: list
        :return: List of :class:`~aquaduct.traj.paths.GenericPathTypeCodes`.
        '''
        # get proper types
        types_dist_orig = self.types_distribution()
        types_dist_range = list(set(types_prob))
        types_thresholds = []
        for t in types_dist_range:
            new_pro_types = [{True: GenericPathTypeCodes.scope_name,
                              False: GenericPathTypeCodes.object_name}[typ >= t] for typ in types_prob]
            types_thresholds.append(make_default_array(cdist(np.matrix(self.simple_types_distribution(new_pro_types)),
                                          types_dist_orig, metric='euclidean')))
            self.beat()
        # get threshold for which value of types_thresholds is smallest
        types = [{True: GenericPathTypeCodes.scope_name,
                  False: GenericPathTypeCodes.object_name}[typ >= types_dist_range[np.argmin(types_thresholds)]] for typ
                 in types_prob]
        return types

    def get_master_path(self, smooth=None, resid=0):
        '''
        .. _master_path_generation:

        Averages spaths into one master path.

        This is done in steps:

        #. Master path is an average of bunch of spaths. Its length is determined by :meth:`full_size` method.
        #. All spaths are then divided in to chunks according to :func:`~aquaduct.utils.helpers.xzip_xzip` function with :attr:`N` set to lenght of master path. This results in list of length equal to the length of master path. Elements of this lists are slice objects that can be used to slice spaths in appropriate chunks.
        #. Next, for each element of this list :meth:`CTypeSpathsCollectionWorker.coords_types_prob_widths` method is called. Types probabilities are changed to types wiht :meth:`types_prob_to_types`.
        #. Finally, all data are used to create appropriate :class:`MasterPath`. If this fails `None` is returned.

        :param Smooth smooth: Smoothing method.
        :param int resid: Residue ID of master path.
        :rtype: :class:`~aquaduct.traj.paths.MasterPath`
        :return: Average path as :class:`~aquaduct.traj.paths.MasterPath` object or `None` if creation of master path failed.
        '''
        # prepare worker
        worker = CTypeSpathsCollectionWorker(spaths=self.spaths, ctype=self.ctype, bias_long=self.bias_long,
                                             smooth=smooth)
        # add some spaths precalcualted properties to worker
        worker.lens_cache = self.lens_cache
        worker.lens_real_cache = self.lens_real_cache
        worker.lens_norm_cache = self.lens_norm_cache
        worker.full_size_cache = self.full_size_cache

        # desired full size of path
        full_size = self.full_size_cache

        # containers for coords, types and widths of master path
        coords = [None] * full_size
        types = [None] * full_size
        widths = [None] * full_size

        # pbar magic
        pbar_previous = 0
        pbar_factor = float(len(self.spaths)) / full_size

        # create pool of workers - mapping function
        map_fun = map
        if self.threads > 1:
            pool = multiprocessing.Pool(self.threads)
            map_fun = pool.imap_unordered
            chunk_size = int(full_size / self.threads ** 2)
            if chunk_size == 0:
                chunk_size = 1
            map_fun = partial(pool.imap_unordered,chunksize=chunk_size)

        # TODO: it is possible to add pbar support here!
        # maximal number of spath
        spath_nr_max = 0
        # loop over results of workers calculations on xzip_xzip lens_real with N=full_size
        # 1. Lens_real (sizes of spaths) are submitted to xzip_xzip wih N=full size
        #    For each spath there will be collection of N slices, each slice cuts some part of spath.
        #    In consequence, all spaths will be cutted in tho N chunks and for each path chunk will be
        #    of different size
        # 2. These slices are submitted to worker callable class.
        for pbar_nr, (spath_nr, (coords_, types_, widths_)) in enumerate(
                map_fun(worker, enumerate(xzip_xzip(*worker.lens_real_cache, N=full_size)))):
            coords[spath_nr] = coords_
            types[spath_nr] = types_
            widths[spath_nr] = widths_
            spath_nr_max = max(spath_nr, spath_nr_max)
            pbar_current = int((pbar_nr + 1) * pbar_factor)
            if pbar_current > pbar_previous:
                pbar_previous = pbar_current
                self.update()  # update progress bar
            else:
                self.beat()
        assert pbar_nr == spath_nr_max, "Internal error. Final global progress of master path generation not synced with maximal number of spath. Please send a bug report to developer(s): %s" % clui.mail

        if self.threads > 1:
            pool.close()
            pool.join()
            pool.terminate()
            del pool

        # at this stage we have coords, widths and types probability

        # get proper types
        with clui.tictoc('proper tests in %s' % str(self.ctype)):
            types = self.types_prob_to_types(types)

        # make frames
        frames = range(len(coords))

        # finalize

        # max min frames
        min_pf = 0
        max_pf = len(coords) - 1
        if self.ctype is None:  # this never happens because of assertion in __init__
            min_pf = None
            max_pf = None
        else:
            if self.ctype.input is not None:
                min_pf = None
            if self.ctype.output is not None:
                max_pf = None

        with clui.tictoc('generic paths in %s' % str(self.ctype)):
            # get and populate GenericPath
            gp = GenericPaths(resid, min_pf=min_pf, max_pf=max_pf)
            for c, t, f in zip(coords, types, frames):  # TODO: remove loop
                gp.add_type(f, t)
                gp.add_coord(c)
        # now try to get first SinglePath, if unable issue WARNING
        with clui.tictoc('separate paths in %s' % str(self.ctype)):
            try:
                sp = list(yield_single_paths([gp]))[0]
            except IndexError:
                logger.warning('No master path found for ctype %s' % str(self.ctype))
                return None
        # finally get MasterPath and add widths
        mp = MasterPath(sp)
        mp.add_width(widths)
        return mp


'''
def create_master_spath(spaths, smooth=None, resid=0, ctype=None, bias_long=5, pbar=None):
    def beat():
        if pbar is not None:
            pbar.heartbeat()

    def update():
        if pbar is not None:
            pbar.update(1)

    # first check what is the size of paths in all parts and normalize and then scale them
    sizes = []
    for part in parts:
        # lengths of all paths of part part
        lens = make_default_array([float(len(sp.types[part])) for sp in spaths])
        if np.max(lens) > 0:
            lens /= np.max(lens)  # normalization
            lens = lens ** bias_long  # scale them by increasing weights of long paths
        if sum(lens) == 0:
            sizes.append(0)
        else:
            # weighted average by paths lengths
            sizes.append(int(np.average([len(sp.types[part]) for sp in spaths], 0, lens)))
        beat()
    full_size = sum(sizes)  # total size (desired)
    pbar_factor = float(len(spaths)) / full_size

    # get total lenghts of all paths - needed as weights in averaging
    lens = make_default_array([float(sp.size) for sp in spaths])
    # if ctype in #:# and not 0 and not None then take object part only length
    if ctype is not None:
        if ctype.input is not None:
            if ctype.input > 0:
                if ctype.input == ctype.output:
                    lens = make_default_array([float(len(sp.types_object)) for sp in spaths])
    # normalize and incearse weight of long paths (depends ob ctype)
    if np.max(lens) > 0:
        lens /= np.max(lens)  # normalize
        lens = lens ** bias_long  # bias to long paths

    # containers for coords, types and widths of master path
    coords = []
    types = []
    widths = []
    pbar_previous = 0

    # following loop calculates types distribution only
    # loop over zip zipped [smooth] coords of all paths and gtypes with size set to full_size
    for pbar_nr, (coords_zz, types_zz) in enumerate(
            zip(zip_zip(*[sp.get_coords_cont(smooth=smooth) for sp in spaths], N=full_size),
                zip_zip(*[sp.gtypes_cont for sp in spaths], N=full_size))):
        # make lens_zz which are lens corrected to the lenghts of coords_zz and normalized to zip_zip number of obejcts
        lens_zz = []
        for l, coord_z in zip(lens, coords_zz):
            if len(coord_z) > 0:
                lens_zz.append([float(l) / len(coord_z)] * len(coord_z))  # normalize and correct lengths
            else:
                # lens_zz.append([float(l)] * len(coord_z))
                lens_zz.append([])
        # concatenate zip_zip coords and lens
        coords_zz_cat = list(concatenate(*coords_zz))
        lens_zz_cat = list(concatenate(*lens_zz))
        # average coords_zz_cat using weights of lens_zz_cat
        coords.append(make_default_array(np.average(coords_zz_cat, 0, lens_zz_cat)))
        # calculate widths
        if len(coords_zz) > 1:
            # try tu use weighted distance - wminkowski with p=2 is equivalent to weighted euclidean
            # id_of_max = np.argmax(pdist(coords_zz_cat, 'wminkowski', p=2, w=lens_zz_cat))
            # widths.append(pdist(coords_zz_cat, 'euclidean')[id_of_max])
            widths.append(np.mean(pdist(coords_zz_cat, 'euclidean')))
        else:
            widths.append(0.)
        # concatenate zip_zip gtypes
        types_zz_cat = list(concatenate(*types_zz))
        # # pick correct type..., check distance of coords[-1] to coords_zz_cat
        # types_cdist = cdist(np.matrix(coords[-1]), coords_zz_cat, metric='euclidean')
        # types.append(types_zz_cat[np.argmin(types_cdist)])
        # types.append(decide_on_type(Counter(types_zz_cat)))
        types.append(float(types_zz_cat.count(GenericPathTypeCodes.scope_name)) / len(types_zz_cat))

        pbar_current = int((pbar_nr + 1) * pbar_factor)
        if pbar_current > pbar_previous:
            pbar_previous = pbar_current
            update()  # update progress bar
        else:
            beat()
    # get proper types
    # make median distribuitions
    types_dist_orig = np.matrix(np.median([CTypeSpathsCollection.simple_types_distribution(sp.gtypes_cont) for sp in spaths], axis=0),dtype=defaults.float_default)
    types_dist_range = list(set(types))
    types_thresholds = []
    for t in types_dist_range:
        new_pro_types = [{True: GenericPathTypeCodes.scope_name,
                          False: GenericPathTypeCodes.object_name}[typ >= t] for typ in types]
        types_thresholds.append(make_default_array(cdist(np.matrix(CTypeSpathsCollection.simple_types_distribution(new_pro_types)),
                                      types_dist_orig, metric='euclidean')))
        beat()
    # get threshold for which value of types_thresholds is smallest
    types = [{True: GenericPathTypeCodes.scope_name,
              False: GenericPathTypeCodes.object_name}[typ >= types_dist_range[np.argmin(types_thresholds)]] for typ in
             types]

    frames = range(len(coords))

    min_pf = 0
    max_pf = len(coords) - 1

    if ctype is None:
        min_pf = None
        max_pf = None
    else:
        if ctype.input is not None:
            min_pf = None
        if ctype.output is not None:
            max_pf = None

    gp = GenericPaths(resid, min_pf=min_pf, max_pf=max_pf)
    for c, t, f in zip(coords, types, frames):  # TODO: remove loop
        gp.add_type(f, t)
        gp.add_coord(c)
        beat()
    # return gp
    beat()  # touch progress bar
    try:
        sp = list(yield_single_paths([gp]))[0]
    except IndexError:
        logger.warning('No master path found for ctype %s' % str(ctype))
        return None
    beat()  # touch progress bar
    mp = MasterPath(sp)
    mp.add_width(widths)
    return mp
'''

'''
def calculate_master(spaths_resid_ctype_smooth):
    spaths, resid, ctype, smooth = spaths_resid_ctype_smooth
    return CTypeSpathsCollection(spaths=spaths, ctype=ctype).get_master_path(smooth=smooth, resid=resid)

'''



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg


class Center(object):
    def __init__(self, X):
        self.mean = np.mean(X, 0)

    def __call__(self, X):
        return X - self.mean

    def undo(self, X):
        return X + self.mean


class Normalize(object):
    def __init__(self, X):
        self.std = np.std(X, 0)

    def __call__(self, X):
        return X / self.std

    def undo(self, X):
        return X * self.std


class Standartize(Center, Normalize):
    def __init__(self, X):
        Center.__init__(self, X)
        Normalize.__init__(self, X)

    def __call__(self, X):
        return Normalize.__call__(self, Center.__call__(self, X))

    def undo(self, X):
        return Center.undo(self, Normalize.undo(self.X))


class PCA(object):
    def __init__(self, X, prepro=None):
        self.preprocess_method = prepro
        # SVD
        self.U, self.d, self.Pt = scipy.linalg.svd(self.preprocess(X), full_matrices=False)
        assert np.all(self.d[:-1] >= self.d[1:]), "SVD error."  # sorted
        # PCA
        self.T = self.U * self.d
        self.eigen = self.d ** 2
        self.sumvariance = np.cumsum(self.eigen)
        self.sumvariance /= self.sumvariance[-1]

    @property
    def P(self):
        return self.Pt.T

    def preprocess(self, X):
        if self.preprocess_method is None:
            return X
        return self.preprocess_method(X)

    def preprocess_undo(self, X):
        if self.preprocess_method is None:
            return X
        return self.preprocess_method.undo(X)

    def __call__(self, X):
        return np.dot(self.preprocess(X), self.P)

    def undo(self, T):
        return self.preprocess_undo(np.dot(T, self.Pt))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from setuptools import find_packages, setup

version = __import__('aquaduct').version_nice()  # version str

install_requires =['numpy>=1.7.0',
                  'scipy>=0.14.0',
                  'scikit-learn>=0.16.0',
                  'MDAnalysis[amber]==0.15.0',
                  'roman>=2.0.0',
                  ]

def install_requires_nice(level=0):
      import re
      for ir in install_requires:
            print ("    "*level) + "* " + " ".join(re.split('(>=|==|<=|>|<|=)',ir))

author = __import__('aquaduct').__author__  # version str


setup(name='aquaduct',
      version=version,
      description='Tracing residues in MD simulation',
      author=author,
      author_email='info@aquaduct.pl',
      url='http://aquaduct.pl',
      packages=find_packages(include=['aquaduct*']),
      install_requires=install_requires,
      extras_require={'full_pymol': ["pymol>=1.4"],
                      'graphs': ['matplotlib'],
                      },
      scripts=['apps/valve.py', 'apps/valve_run'],
      provides=['aquaduct'],
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'Operating System :: POSIX',
                   'Programming Language :: Python :: 2.7',
                   ],
      )



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module comprises convieniences functions and definitios for different operations related to command line user interface.
"""

import logging
logger = logging.getLogger(__name__)

from collections import OrderedDict
from aquaduct import logger as root_logger
from aquaduct import __mail__ as mail
import datetime
import time
from os import linesep
from sys import stderr
from os import linesep
from functools import partial

from multiprocessing import Queue, Manager, Lock, Value, Process
from collections import OrderedDict

def emit_message_to_file_in_root_logger(mess):
    # emits message to the file used by file handler in the root logger
    # assumes there is only one file handler
    if logging.FileHandler in map(type, root_logger.handlers):
        fh = root_logger.handlers[map(type, root_logger.handlers).index(logging.FileHandler)]
        with fh.lock:
            with open(fh.baseFilename, 'a') as logfile:
                logfile.write(mess)


def message_special(mess):
    emit_message_to_file_in_root_logger(mess)


def message(mess, cont=False):
    """
    Prints message to standard error.
    If FileHandler is present in the :class:`root_logger` the same message is appended to the log file.

    :param str mess: message to print
    :param bool cont: if set True no new line is printed
    """
    if cont:
        mess = mess + ' '
    else:
        mess = mess + linesep
    emit_message_to_file_in_root_logger(mess)
    stderr.write(mess)


class fbm(object):
    # feedback message
    def __init__(self, info, cont=True):
        self.__cont = cont
        self.__info = info
        message(info + '...', cont=self.__cont)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        if typ is None:
            if self.__cont:
                message("OK.")
            else:
                message(self.__info + ": DONE.")

    def __call__(self, info):
        if self.__cont:
            message(linesep + '\t' + info + '...', cont=True)
        else:
            message(info + '...', cont=False)


class tictoc(object):
    def __init__(self, mess):
        self.__tic = 0
        self.__toc = 0
        self.__mess = mess

    def __enter__(self):
        self.__tic = time.time()
        return self

    def __exit__(self, typ, value, traceback):
        self.__toc = time.time()
        if typ is None:
            #logger.debug('Execution time of [%s] %s', self.__mess, smart_time_string(self.__toc - self.__tic))
            logger.debug('Execution time of [%s] %f', self.__mess, (self.__toc - self.__tic))


gregorian_year_in_days = 365.2425
'''Length of Gregorian year in days. Average value. Source: https://en.wikipedia.org/wiki/Year'''


def smart_time_string(s, rl=0, t=1.1, maximal_length=None, maximal_units=5):
    """
    Function transforms time in seconds to nicely formatted string of
    length defined by :attr:`maximal_length`. Depending on number of seconds time is represented with
    one or more of the following units:

    ========= =================
    Unit name Unit abbreviation
    ========= =================
    seconds   s
    minutes   m
    hours     h
    days      d
    years     y
    ========= =================

    Maximal number of units used in time string can be set with :attr:`maximal_units`.

    :param int s: Input time in seconds.
    :param int rl: Number of units already used for representing time.
    :param float t: Exces above standard number of current time units.
    :param int maximal_length: Maximal length of the output string. Must be greater then 0.
    :param int maximal_units: Maximal number of units used in the output string. Must be greater then 0 and lower then 6.

    :return: string of nicely formated time
    :rtype: str
    """
    # assert isinstance(maximal_length, (int, long))
    # assert maximal_length > 0
    assert isinstance(maximal_units, (int, long))
    assert maximal_units > 0
    assert maximal_units < 6

    output = ''
    rl += 1
    if rl > maximal_units:
        output = ''
    else:
        # seconds
        if s < t * 60:
            output = "%2.2d s" % s
        elif s < t * 3600:
            output = ("%2.2d m" % (int(s) / 60)) + ' ' + smart_time_string(int(s) % 60, rl).strip()
        elif s < t * 3600 * 24:
            output = ("%2.2d h" % (int(s) / 3600)) + ' ' + smart_time_string(int(s) % 3600, rl).strip()
        elif s < t * 3600 * 24 * gregorian_year_in_days:
            output = ("%d d" % (int(s) / (3600 * 24))) + ' ' + smart_time_string(int(s) % (3600 * 24), rl).strip()
        elif True:
            output = ("%d y" % (int(s) / (3600 * 24 * gregorian_year_in_days))) + ' ' + smart_time_string(
                int(s) % (3600 * 24 * gregorian_year_in_days), rl).strip()
    if maximal_length:
        return (output + " " * maximal_length)[:maximal_length]
    return output


###################################
# vis separators

def gsep(sep='-', times=72, length=None):
    """
    Generic separator.

    :param str sep: Element(s) of separator.
    :param int times: Number of times :attr:`sep` is printed.
    :param int length: Optional maximal length of output.
    :return: String separator.
    :rtype: str
    """
    return (sep * times)[:length]


def tsep(line):
    """
    :param str line: Input line.
    :return: Returns default :func:`gsep` of length of :attr:`line`.
    """
    return gsep(sep='-', times=len(line))


def underline(line):
    """
    :param str line: Input line.
    :return: String made by concatenation of :attr:`line`, :mod:`os.linesep`, and output of :func:`tsep` called with :attr:`line`.
    :rtype: str
    """
    uline = line
    uline += linesep
    uline += tsep(line)
    return uline


def thead(line):
    """
    :param str line: Input line.
    :return: String made by concatenation of output of :func:`tsep` called with :attr:`line`, :attr:`line`, :mod:`os.linesep`, and again output of :func:`tsep` called with :attr:`line`.
    :rtype: str
    """
    header = tsep(line)
    header += linesep
    header += line
    header += linesep
    header += tsep(line)
    return header


###################################


class SimpleProgressBar(object):
    """
    Simple progress bar displaying progress with percent indicator, progress bar and ETA.
    Progress is measured by iterations.

    :cvar str rotate: String comprising characters with frames of a rotating toy.
    :cvar int barlenght: Length of progress bar.
    :ivar int maxval: maximal number of iterations
    :ivar int current: current number of iterations
    :ivar bool overrun_notice: if True, overrun above :attr:`maxval` iterations causes insert of newline
    :ivar bool overrun: flag of overrun
    :ivar int begin: time in seconds at the initialization of the :class:`SimpleProgressBar` class.
    :ivar int tcurrent: time in seconds of current iteration
    """

    rotate = '\\|/-'
    # rotate = '<^>v'
    # rotate = '.:|:.'
    # rotate = 'x+'

    barlenght = 24

    def __init__(self, maxval=None, mess=None):
        """
        :param int maxval: Maximal number of iterations stored to :attr:`maxval`.
        :param str mess: Optional message displayed at progress bar initialization.
        """

        #self.lock = Manager().Lock()

        assert isinstance(maxval,
                          (int, long)), 'Parameter maxval should be of int or long type, %r given instead.' % type(
            maxval)
        if maxval < 1:
            self.maxval = 1
        else:
            self.maxval = maxval

        self.tens = []

        self.current = 0

        self.overrun_notice = True
        self.overrun = False

        self.begin = time.time()
        self.tcurrent = self.begin

        self.last_rotate_time = self.begin
        self.last_rotate_idx = 0

        if mess is not None:
            message(mess)
        self.show()

    def bar(self):
        barval = int(self.percent() / 100 * self.barlenght)
        if barval > self.barlenght:
            barval = self.barlenght
        bar = '#' * barval
        if self.current:
            if self.tcurrent - self.last_rotate_time > 1. / 4:  # FIXME: magic constant, remove it!
                self.last_rotate_idx += 1
                self.last_rotate_time = self.tcurrent
            if self.last_rotate_idx > len(self.rotate) - 1:
                self.last_rotate_idx = 0
            bar += self.rotate[self.last_rotate_idx]
        bar += ' ' * self.barlenght
        return '[%s]' % bar[:self.barlenght]

    def ETA(self):
        """
        Returns ETA calculated on the basis of current number of iterations
        :attr:`current` and current time :attr:`tcurrent`. If number of
        iterations is 0 returns ``?``.
        Time is formated wiht :func:`smart_time_string`.

        :return: ETA as string.
        :rtype: str
        """
        if self.current == 0:
            return '?'
        diff = self.tcurrent - self.begin
        periteration = diff / self.current
        expected = periteration * self.maxval
        eta = periteration * (self.maxval - self.current)

        return smart_time_string(eta)

    def percent(self):
        """
        Returns float number of precent progress calculated in the basis
        of current number of iterations :attr:`current`. Should return
        number between 0 and 100.

        :returns: percent progress number
        :rtype: float
        """
        percent = float(self.current) / float(self.maxval) * 100
        return percent

    def show(self):
        """
        Shows current progress.

        If value returned by :meth:`percent` is =< 100 then progres is printed as
        percent indicator leaded by ETA calculated by :meth:`ETA`.

        If value returned by :meth:`percent` is > 100 then progress is printed as
        number of iterations and total time.

        Progress bar is writen to standard error.
        """
        percent = self.percent()
        # TODO: create some unittests for pbar
        mess = ''
        mess_spec = ''
        if percent > 100 or self.overrun:
            if self.overrun_notice:
                stderr.write(linesep)
                self.overrun_notice = False
                self.overrun = True
            mess_spec = "%d iterations out of %d. Total time: %s" % (self.current, self.maxval, self.ttime())
            mess = "\r" + mess_spec
        elif not self.overrun:
            mess_spec = "%3d%% %s ETA: %s" % (self.percent(), self.bar(), self.ETA())
            mess = "\r" + mess_spec + "\033[K"  # FIXME: magic constant!

        stderr.write(mess)
        # TODO: do not use last_rotate_time here, use separate marker, last_rotate_time can be used in over run notice
        if int(percent) / 10 not in self.tens:  # FIXME: magic constant!
            if percent > 100:
                if self.tcurrent - self.last_rotate_time > 60.:  # FIXME: magic constant, remove it!
                    message_special(mess_spec + linesep)
                    self.tens.append(int(percent) / 10)
                    self.last_rotate_time = self.tcurrent
            else:
                message_special(mess_spec + linesep)
                self.tens.append(int(percent) / 10)

    def heartbeat(self):
        #with self.lock:
        if time.time() - self.last_rotate_time > 2.:  # FIXME: magic constant, remove it!
            self.tcurrent = time.time()
            self.show()

    def next(self):
        return self.update(self.current+1)

    def update(self, step):
        """
        Updates number of current iterations :obj:`current` by one if :obj:`step` is > 0.
        Otherwise number of current iterations is not updated.
        In boths cases time of current iteration :obj:`tcurrent` is updated and
        :meth:`show` is called.

        :param int step: update step
        """
        # TODO: change logic of step == 1 vs step > 1 - add or set?
        #with self.lock:
        if step > 0:
            self.current = step
        self.tcurrent = time.time()
        if (step == self.maxval) or (
                        self.tcurrent - self.last_rotate_time > 1. / 4):  # FIXME: magic constant, remove it!
            # TODO: check for last_rotate_time is done twice, SimpleProgressBar code needs revision
            self.show()

    def ttime(self):
        """
        Calculates and returns total time string formated with :func:`smart_time_string`.

        :return: string of total time
        :rtype: str
        """
        return smart_time_string(self.tcurrent - self.begin)

    def finish(self):
        """
        Finishes progress bar. First, :meth:`update` is called with :obj:`step` = 0. Next message of total time
        is writen to standard error.
        """
        if self.current < self.maxval:
            self.update(self.maxval)
        else:
            self.update(0)
            self.show()
        stderr.write(linesep)
        message("Total time: %s" % self.ttime())
        # stderr.write(linesep)


pbar = SimpleProgressBar  # default progress bar


def get_str_timestamp():
    # returns time stamp as string
    return str(datetime.datetime(*tuple(time.localtime())[:6]))



class SimpleTree(object):
    def __init__(self,name=None,message=None):
        self.name = name
        self.message = []
        self.add_message(message)
        self.branches = []

    def __repr__(self):
        return "%s {%s} %s" % (str(self.name), "; ".join(self.message),str(self.branches))

    def is_leaf(self):
        return len(self.branches)==0

    @property
    def leafs_names(self):
        return [leaf.name for leaf in self.branches]

    def get_leaf(self,name):
        assert name in self.leafs_names
        return [leaf for leaf in self.branches if name == leaf.name][0]


    def add_message(self,message=None,toleaf=None,replace=False):
        if toleaf is not None:
            return self.add_message_to_leaf(message=message,toleaf=toleaf,replace=replace)
        if message is not None:
            if isinstance(message,list):
                if replace:
                    self.message = message
                else:
                    self.message += message
            else:
                if replace:
                    self.message = [message]
                else:
                    self.message += [message]

    def add_message_to_leaf(self,message=None,toleaf=None,replace=False):
        if toleaf in self.leafs_names:
            leaf = self.get_leaf(toleaf)
            return leaf.add_message(message,replace=replace)
        else:
            for leaf in self.branches:
                leaf.add_message_to_leaf(message=message,toleaf=toleaf,replace=replace)

    def add_leaf(self,name=None,message=None,toleaf=None):
        if toleaf is not None:
            return self.add_leaf_to_leaf(name=name,message=message,toleaf=toleaf)
        leaf = SimpleTree(name=name,message=message)
        self.branches.append(leaf)

    def add_leaf_to_leaf(self,name=None,message=None,toleaf=None):
        if toleaf in self.leafs_names:
            leaf = self.get_leaf(toleaf)
            return leaf.add_leaf(name=name,message=message)
        else:
            for leaf in self.branches:
                leaf.add_leaf_to_leaf(name=name,message=message,toleaf=toleaf)


def print_simple_tree(st,prefix=None,multiple=False,concise=True):
    _l = '|'
    _t = ' '
    _c = '-+'+_t

    def name_str(name):
        if name is None:
            return ''
        return str(name)
    def name_len(name):
        return len(name_str(name))

    def message_str(message):
        if len(message):
            return '{%s}' % '; '.join(message)
        return ''

    prefix_ = ''
    if prefix is not None:
        prefix_ += prefix
    out = ''
    # name
    out += prefix_
    out += name_str(st.name)
    if not st.is_leaf():
        if name_len(st.name):
            out += _c
        out += message_str(st.message)
        out += linesep
        if multiple:
            new_prefix = prefix_+_l+(_t*(name_len(st.name)-1))
        else:
            new_prefix = prefix_+(_t*(name_len(st.name)))
        new_prefix += _t
        if not concise:
            out += new_prefix+_l+linesep
        out_rec = []
        for nr,branch in enumerate(st.branches):
            out_rec.append(partial(print_simple_tree,prefix=new_prefix, multiple=len(st.branches) - nr > 1)(branch))
        out += ''.join(out_rec)
        if st.branches[-1].is_leaf() and not concise:
            out += new_prefix.rstrip(_t) + linesep
    else:
        out += _t
        out += message_str(st.message)
        out += linesep
    return out



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.spatial.distance import pdist


################################################################################
# traces utils and helpers

def diff(trace):
    """
    This function calculates the distance between 2 given points.

    :param trace: coordinates in numpy array object
    :return: distance between points
    """
    # trace - 3D coordinates
    # returns distances between coordinates
    assert isinstance(trace, np.ndarray), "Trace should be of np.ndarray type, %r submited instead." % type(trace)
    assert len(trace.shape) == 2, "Traces should be 2d, %dd submited instead (trace no. %d)" % len(trace.shape)

    n = len(trace)
    if n < 2:
        return None

    trace_diff = []
    for nr, row in enumerate(trace[:-1]):
        trace_diff.append(float(pdist(np.vstack((row, trace[nr + 1])), metric='euclidean')))

    return np.array(trace_diff)


def tracepoints(start, stop, nr):
    """
    :param start: coordinates of the first point as a numpy array object
    :param stop: coordinates of the second point as a numpy array object
    :param nr: number of elements between the first and second point
    :return: two-dimentional numpy array; number of dimentions depends on nr parameter
    """
    # returns points between start and stop as linear interpolations
    # if nr == 1 then midpoint is returned
    return np.array([np.linspace(cb, ce, nr + 2)[1:-1] for cb, ce in zip(start, stop)]).T


def midpoints(paths):
    """
    The function returns a tuple of numpy arrays extended with mid point spanning last and first element(column)
     of these arrays.

    :param paths: a tuple of 2-dimentional np.arrays that hold 3D coordinates; each element holds one trace, all elements are supposed to make one path divided in to sections
    :return: paths elements with additional mid points as a generator object
    """
    # paths - a tuple of 2d np.arrays that hold 3D coordinates, each element holds one trace, all elements are supposed to make one path divided in to sections
    # yields paths elements with additional mid points
    # if input paths is follwoing:
    #   11111 33333 55555
    # function yields the same elements plus midpoints:
    #   111112 2333334 455555
    assert isinstance(paths, tuple), "Paths should be of tuple type, %r submitted instead." % type(paths)
    for nr, trace in enumerate(paths):
        assert isinstance(trace,
                          np.ndarray), "Traces should be of numpy.ndarray type, %r submited instead (trace no. %d)" % (
            type(trace), nr)
        # assert len(trace.shape) == 2, "Traces should be 2d, %dd submited instead (trace no. %d)" %(len(trace.shape),nr)

    n = len(paths)
    last_trace = None
    if n > 1:
        for nr, trace in enumerate(paths):
            # find past and next trace
            past = []
            if nr - 1 >= 0:
                if paths[nr - 1].size > 0:
                    past = paths[nr - 1][-1]
            next_trace = []
            if nr + 1 < n:
                if paths[nr + 1].size > 0:
                    next_trace = paths[nr + 1][0]
            # calculate midpoints if relevant
            if len(trace) > 0:
                if len(past) > 0:
                    midp = tracepoints(past, trace[0], 1)
                    trace = np.vstack((midp, trace))
                if len(next_trace) > 0:
                    midp = tracepoints(trace[-1], next_trace, 1)
                    trace = np.vstack((trace, midp))
            yield trace
    else:
        yield paths[0]


def length_step_std(trace):
    """
    This function calculates sum, mean and standard deviation from all segments of a trace.

    :param trace: coordinates of points as numpy array
    :return: a tuple with basics statistics of a trace
    """
    # trace - 3D coordinates
    # calculates diff over trace and returns sum, mean and std of diff
    # if trace is empty or have length < 2 nans are returned
    if len(trace) < 2:
        return float('nan'), float('nan'), float('nan')
    d = diff(trace)
    return np.sum(d), np.mean(d), np.std(d)


def derrivative(values):
    # FIXME: This function is porbably completely wrong! AP is going to rewrite it.
    # values - 3D coordinates
    # calculates derrivative of lenght of trace
    # uses diff but yields the same number of values as in input data
    # this is done by interpolation and applying simple correction
    diff = np.diff(values)
    size = len(diff)
    correction = 1. / (size + 1)  # this correct values so after integration they are closer to expected value
    # This was calculated by following experiment:
    # w = []
    # for q in range(10000):
    #     r = np.cumsum(np.random.rand(10000))
    #     w.append(r[-1]/sum(list(traces.derrivative(r))))
    # np.mean(w) # w is of narmal distribution

    for nr in range(size + 1):
        if nr == 0:
            yield diff[0] - correction  # begin
        elif nr == size:
            yield diff[-1] - correction  # end
        else:
            yield (diff[nr - 1] + diff[nr]) / 2. - correction

# todo : aby zaoszczedzic na obliczeniach mozna pomijac takie katy(lub zwracac 0), ktorych zwracane wartosci są bardzo,bardzo małe (rzedu np 10**-4)-> np kat 0.005 rad to 0,29stopnia miary łukowej
# wektory: promień atomu wodoru to 0.529A

################################################################################
# vectors, triangles and angles helpers

def vector_norm(V):
    """

    :param V: a vector in a form of array-like object, tuple or a list
    :return: normalized length of a vector
    """
    # calculate length of physical vector based on it's coordinates
    # input: tuple or a list
    # output: float
    # return np.sqrt(np.dot(V, V.conj()))
    return np.sqrt(np.dot(V, V))
    # return np.linalg.norm(V)


def triangle_angles(A, B, C):
    """
    Parameters are coordinates of points which are tops of triangle. The function calculates angles in a triangle formed by given coordinates.

    :param A: coordinates of the first point
    :param B: coordinates of the second point
    :param C: coordinates of the third point
    :return: list of arguments where angle is given in radians , the output is as follow: [BAC,CAB,ABC]
    """
    # http://stackoverflow.com/questions/5122372/angle-between-points
    # A,B,C are point in the space
    # input: 3 space coords of points (as tuple or list)
    # returns list of arguments where angle is given in radians , the output is as follow: [BAC,CAB,ABC]
    A, B, C = map(np.array, (A, B, C))
    a = C - A
    b = B - A
    c = C - B
    angles = []
    for e1, e2 in ((a, b), (a, c), (b, -c)):
        num = np.dot(e1, e2)
        denom = vector_norm(e1) * vector_norm(e2)
        angles.append(np.arccos(np.clip(num / denom, -1, 1)))  # cliping values in to [-1,1]
        if np.isnan(angles[-1]):
            angles[-1] = 0.
    return angles


def triangle_angles_last(A, B, C):
    """
    Parameters are coordinates of points which are tops of triangle. The function calculates the [ABC] angle.

    :param A: coordinates of the first point [A top]
    :param B: coordinates of the second point [B top]
    :param C: coordinates of the third point [C top]
    :return: list with one value of ABC angle in radians
    """
    # http://stackoverflow.com/questions/5122372/angle-between-points
    # A,B,C are point in the space
    # input: 3 space coords of points (as tuple or list)
    # returns list with one value of ABC angle in radians
    A, B, C = map(np.array, (A, B, C))
    a = C - A
    b = B - A
    c = C - B
    angles = []
    for e1, e2 in ((b, -c),):
        num = np.dot(e1, e2)
        denom = vector_norm(e1) * vector_norm(e2)
        angles.append(np.arccos(np.clip(num / denom, -1, 1)))  # cliping values in to [-1,1]
    return angles


def triangle_height(A, B, C):
    """
    Parameters are coordinates of points which are tops of triangle. The function calculates the ABC triangle height.

    :param A: coordinates of the first point [A top]
    :param B: coordinates of the second point [B top]
    :param C: coordinates of the third point [C top]
    :return: one value of ABC triangle height
    """
    # a is head
    # input: 3 space coords of points (as tuple or list)
    # output float, value of triangle height
    angles = triangle_angles_last(A, B, C)  # ta funkcja zwraca tylko 1 element
    A, B, C = map(np.array, (A, B, C))
    c = vector_norm(B - A)
    h = np.sin(angles[-1]) * c
    if np.isnan(h):
        h = 0.
    return h


def vectors_angle(A, B):
    """
    This function calculates the angle between two given vectors (starting from the [0,0,0] to the given coordinates.

    :param A: coordinates of the first point which is the end of the vector
    :param B: coordinates of the second point which is the end of the vector
    :return: the angle between vectors in question (in radians)
    """
    angle = np.arccos(np.dot(A, B) / (vector_norm(A) * vector_norm(B)))
    if np.isnan(angle):
        return 0.0
    return angle


def vectors_angle_alt(A, B):
    '''
    This function calculates the angle between two given vectors (starting from the [0,0,0] to the given coordinates

    * alternative method.

    :param A: coordinates of the first point which is the end of the vector
    :param B: coordinates of the second point which is the end of the vector
    :return: the angle between vectors in question (in radians)
    '''
    return np.arccos(np.clip(np.dot(A / vector_norm(A), B / vector_norm(B)), -1.0, 1.0))


def vectors_angle_alt_anorm(A, B, A_norm):
    """
    This function calculates the angle between two given vectors (starting from the [0,0,0] to the given coordinates
     - alternative method with additional A_norm holding norm of A.

    :param A: coordinates of the first point which is the end of the vector
    :param B: coordinates of the second point which is the end of the vector
    :param A_norm: additional parameter holding normalized of vector A
    :return: the angle between vectors in question (in radians)
    """
    return np.arccos(np.clip(np.dot(A / A_norm, B / vector_norm(B)), -1.0, 1.0))


def vectors_angle_anorm(A, B, A_norm):
    """
    This function calculates the angle between two given vectors (starting from the [0,0,0] to the given coordinates
     using additional A_norm holding norm of A.

    :param A: coordinates of the first point which is the end of the vector
    :param B: coordinates of the second point which is the end of the vector
    :param A_norm: additional parameter holding normalized of vector A
    :return: the angle between vectors in question (in radians)
    """
    norm2 = A_norm * vector_norm(B)
    if norm2 == 0.:
        return 0.
    angle = np.clip(np.dot(A, B) / norm2, -1., 1.)
    if np.isnan(angle):
        return 0.
    return np.arccos(angle)


################################################################################
# Linearization classes

# poprawiony algorytm
class LinearizeOneWay(object):
    def here(self, coords):
        # coords - 3D coordintates of a trace
        # yields indices of coords which is a staring point of linear fragments of the trace and the next point after
        # enf of linear segment; done in one way
        """
        This function simplifies the trace by removing the redundant, linear points
        :param coords: 3D coordinates of a trace as an array-like object
        :return: indices of coordinates which are a staring and ending points of linear fragments and other non-linear
        points of the trace
        """
        size = len(coords)
        yield 0
        if size <= 3:
            yield 2
        else:
            for sp in range(size - 2):
                ep = sp + 3
                if self.is_linear(coords[sp:ep]):
                    continue
                yield sp + 1
                continue
            yield size - 1


class LinearizeHobbit(LinearizeOneWay):
    def and_back_again(self, coords):
        # coords - 3D coordintates of a trace
        # yields indices of coords that spans linear fragments of the trace; done in opposite way than in one way
        size = len(coords)
        return (size - e - 1 for e in self.here(coords[::-1]))

    def __call__(self, coords):
        # coords - 3D coordintates of a trace
        # wrapper that uses here and and_back_again methods to get merged uniq and sorted indices of coords that spans linear fragments of the trace
        # returns these points from coords that are linear simplification of coords
        # __call__ is required by child classes
        here = self.here(coords)
        and_back_again = self.and_back_again(coords)
        linearize = sorted(list(set(list(here) + list(and_back_again))))
        return coords[linearize]


class LinearizeRecursive(object):
    """
    Base class for linearization methods classes.

    It implements recursive algorithm.
    """

    def here(self, coords, depth=0):
        """
        Core of recursive linearization argorithm.

        It checks if the first, the last and the middle point are linear according to the criterion. The middle point is a selected point that is in the middle of length of the paths made by input coordinates.

        If these points are linear their indices are returned. Otherwise, coordinates are split into two parts. First part spans points from the first point to the middle point (inclusive) and the second part spans points from the middle (inclusive) to the last point. Next, these two parts are submitted recursively to :meth:`here`.

        Results of these recursive calls are joined, redundant indices are removed and sorted result is returned.

        :param numpy.ndarray coords: Input coordinates.
        :param int depth: Depth of recurence.
        :return: Indices of ``coords`` points that can be used instead of all points in visulatization.
        :rtype: list of int
        """
        # klasa nie ma zdefiniowanej metody is_linear
        depth += 1

        lengths = np.hstack(([0], np.cumsum(diff(coords))))
        size = len(lengths)
        if size <= 3:
            return range(size)
        sp = 0
        ep = size - 1
        mp = np.argwhere(lengths > max(lengths) / 2)
        if len(mp): # in case lenghs are zeros
            mp = int(mp[0])
        else:
            mp = (sp + ep) / 2

        if mp == sp:
            mp += 1
        if mp == ep:
            mp -= 1

        if self.is_linear(coords[[sp, mp, ep]], depth=depth):
            return [sp, mp, ep]
        return sorted(
            list(set(self.here(coords[:mp + 1], depth=depth) + [e + mp for e in self.here(coords[mp:], depth=depth)])))

    def __call__(self, coords):
        # returns these points from coords that are linear simplification of coords
        # __call__ is required by child classes
        here = self.here(coords)
        return coords[here]


################################################################################
# Linearize criteria

class TriangleLinearize(object):
    def __init__(self, threshold=0.01):
        # threshold - maximal allowed sum of heights of triangles made of beginning, end and all middle points
        ## bardzo ostre kryterium!!
        self.threshold = threshold

    def is_linear(self, coords, **kwargs):
        # coords - 3D coordintates of a trace
        # returns True if coords make a straight line
        # criterion of linearity:
        # if sum of heights of triangles made of beginning, end and all middle points does not exceed threshold coords are linear
        list_of_h = list()
        for head in coords[1:-1]:
            list_of_h.append(triangle_height(head, coords[0], coords[-1]))
            # print list_of_h, sum(list_of_h)
            if sum(list_of_h) > self.threshold:
                return False
        return True


class VectorLinearize(object):
    """
    Base class for linearization methods classes.

    It implements vector linearization criterion.
    """

    def __init__(self, treshold=0.05236):
        # TODO: Add docs
        self.treshold = treshold

    def is_linear_core(self, coords, depth=None):
        """
        Method checks if input coordinates are linear according to the threshold and depth.

        It begins with calculation of the threshold. If `depth` is None it is set to 1. Current threshold is calculated with following simple equation:

        .. math::

            threshold_{current} = threshold_{initial} * (2 - 0.9^{depth})

        Next, in a loop over all points but the first and the last the angle is calculated between two vectors. The first one made by the point and the first point, and the second vector made by the last and the first point. If any of the calculated angles is bigger the the treshold methods returns False; otherwise method returns True.

        :param numpy.ndarray coords: Coordinates for which linearization criterion is checked.
        :param int depth: Depth of recurence.
        :return: True if input coordinates are linear and False otherwise.
        :rtype: bool
        """
        if depth is None:
            depth = 1

        treshold = self.treshold + self.treshold * (1 - 0.9 ** depth)  # FIXME: magic constant!

        V = coords[-1] - coords[0]
        V_norm = vector_norm(V)
        for cp in coords[:-1]:
            V_sum = cp - coords[0]
            if vectors_angle_anorm(V, V_sum, V_norm) > treshold:
                return False
        return True

    def is_linear(self, coords, depth=None, **kwargs):

        """
        For more detail see :meth:`is_linear_core` which is used as the criterion of linearity in this method.

        :param numpy.ndarray coords: Coordinates for which linearization criterion is checked.
        :param int depth: Depth of recurence.
        :return: True if input coordinates are linear and False otherwise. Criterion is checked for coordinates in normal and reverse order.
        :rtype: bool
        """
        if not self.is_linear_core(coords, depth=depth):
            return False
        elif not self.is_linear_core(coords[::-1], depth=depth):
            return False
        return True


################################################################################
# Concrete classes for linearization

class LinearizeRecursiveVector(LinearizeRecursive, VectorLinearize):
    """
    ..  _simply_smooths_details:

    Class provides recursive linearization of coordinates with :class:`LinearizeRecursive` algorithm and the criterion of linearity implemented by :class:`VectorLinearize`. This is default method.
    """
    pass


class LinearizeRecursiveTriangle(LinearizeRecursive, TriangleLinearize):
    """
    Class provides recursive linearization of coordinates with :class:`LinearizeRecursive` algorithm and the criterion of linearity implemented by :class:`TriangleLinearize`.
    """
    pass


class LinearizeHobbitVector(LinearizeHobbit, VectorLinearize):
    """
    Class provides recursive linearization of coordinates with :class:`LinearizeHobbit` algorithm and the criterion of linearity implemented by :class:`VectorLinearize`.
    """
    pass


class LinearizeHobbitTriangle(LinearizeHobbit, TriangleLinearize):
    """
    Class provides recursive linearization of coordinates with :class:`LinearizeHobbit` algorithm and the criterion of linearity implemented by :class:`TriangleLinearize`.
    """
    pass


class LinearizeOneWayVector(LinearizeOneWay, VectorLinearize):
    """
    Class provides recursive linearization of coordinates with :class:`LinearizeOneWay` algorithm and the criterion of linearity implemented by :class:`VectorLinearize`.
    """
    pass


class LinearizeOneWayTriangle(LinearizeOneWay, TriangleLinearize):
    """
    Class provides recursive linearization of coordinates with :class:`LinearizeOneWay` algorithm and the criterion of linearity implemented by :class:`TriangleLinearize`.
    """
    pass

################################################################################
